use std::fmt::Display;
use std::sync::Arc;

use anyhow::Result;
use libpt::log::error;
use libpt::log::info;
use libpt::log::warn;
use minijinja::context;
use minijinja::Environment;
use thiserror::Error;
use warp::http::StatusCode;
use warp::reject;
use warp::reply::Response;
use warp::Filter;

use crate::config::Config;
use crate::vault::VaultRef;

use super::Descriptions;

#[derive(Clone)]
pub struct Service<'tp> {
    vault: VaultRef,
    config: Config,
    env: Environment<'tp>,
    text: Descriptions,
}
impl<'tp> Service<'tp> {
    fn new(
        vault: VaultRef,
        config: Config,
        env: Environment<'tp>,
        text: Descriptions,
    ) -> Arc<Self> {
        Self {
            vault,
            config,
            env,
            text,
        }
        .into()
    }
}

pub async fn serve(text: Descriptions, vault: VaultRef, config: Config) -> Result<()> {
    let mut env = Environment::new();
    env.add_template("index", include_str!("../../data/www/admin.html"))?;

    let service = Service::new(vault, config, env, text);
    let service2 = service.clone();

    let routes = warp::path::end()
        .map(move || service2.clone())
        .and_then(overview)
        .or(warp::path("styles.css").and_then(styles))
        .recover(handle_rejection);

    warp::serve(routes)
        .run(service.config.addr_admin.unwrap())
        .await;

    warn!("exited the admin interface");
    Ok(())
}

#[derive(Debug, Error)]
struct TemplateError(#[from] minijinja::Error);

impl reject::Reject for TemplateError {}
impl Display for TemplateError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "something went wrong with producing this page")
    }
}

async fn overview(serv: Arc<Service<'_>>) -> Result<Box<dyn warp::Reply>, warp::Rejection> {
    let contestants = serv
        .vault
        .contestants()
        .await
        .into_iter()
        .collect::<Vec<_>>();
    let winners = serv.vault.winners().await.into_iter().collect::<Vec<_>>();
    let r = Response::new(
        serv.env
            .get_template("index")
            .map_err(TemplateError::from)?
            .render(context!(
                title => "Wooly-Vault",
                author => env!("CARGO_PKG_AUTHORS"),
                year => "2024",
                challenge_idx => serv.config.challenge,
                challenge_title => serv.text.title(),
                challenge_description => serv.text.description(),
                challenge_hints => serv.text.hints(),
                challenge_solution => serv.text.solution(),
                contestants => serv.vault.contestants().await.iter().collect::<Vec<_>>(),
                winners => serv.vault.winners().await.iter().collect::<Vec<_>>(),
                contestants => contestants,
                winners => winners,
                contestants_amount => contestants.len(),
                winners_amount => winners.len(),
            ))
            .map_err(TemplateError::from)?
            .into(),
    );

    Ok(Box::new(r))
}

async fn styles() -> Result<Box<dyn warp::Reply>, warp::Rejection> {
    let r = Response::new(include_str!("../../data/www/styles.css").to_string().into());
    Ok(Box::new(r))
}

async fn handle_rejection(
    err: reject::Rejection,
) -> Result<impl warp::reply::Reply, std::convert::Infallible> {
    let code;
    let message;
    info!("rejecting: {err:?}");

    if err.is_not_found() {
        code = StatusCode::NOT_FOUND;
        message = "page not found";
    } else if let Some(e) = err.find::<TemplateError>() {
        error!("templating error: {e}");
        code = StatusCode::INTERNAL_SERVER_ERROR;
        message = "could not process data to make a page";
    } else {
        error!("unhandled rejection: {:?}", err);
        code = StatusCode::INTERNAL_SERVER_ERROR;
        message = "unhandled rejection";
    }

    Ok(warp::reply::with_status(message, code))
}
